/*
+--------------------------------------+
|  (C) Jaroslaw Arabas, luty 1999      |
+--------------------------------------+
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "losowe.h"
#include "wewy.h"

LOSOWE Losowe;



void t1Randomize(unsigned s);
void t2Randomize();
void t1Init();
void t2Init();
void t3Init();
double t1Jedn(void);
double t2Jedn(void);
double t3Jedn(void);
double t1Norm();
double t2Norm();
double t3Norm();
double t4Norm();


void Randomize()
{
   switch (Losowe->tInicjacji)
   {
   case 1:  t1Randomize(Losowe->InitZiarno);
            break;
   case 2:  t2Randomize();
            break;
   default: t2Randomize();
            break;
   }
}

double jednostajny ()
{
   switch (Losowe->tJednostajny)
   {
   case 1:  return t1Jedn();
   case 2:  return t2Jedn();
   case 3:  return t3Jedn();
   default: return 0;
   }
}

double normstd()
{
   switch (Losowe->tNormalny)
   {
   case 1:  return t1Norm();
   case 2:  return t2Norm();
   case 3:  return t3Norm();
   case 4:  return t4Norm();
   default: return t1Norm();
   }
}

void initLosowe()
{
   switch (Losowe->tJednostajny)
   {
   case 1:  t1Init();
            break;
   case 2:  t2Init();
            break;   
   case 3:  t3Init();
            break;   
   default: 
            break;
   }
}

void t1Randomize(unsigned s)
{
   srand(s);
}

void t2Randomize()
{
   srand((unsigned)time(0));
}



/* Generator uniwersalny o okresie 2^144, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 39 */

static double uu[97];
static int ip=97;
static int jp=33;
static double cc=362436.0/16777216.0;
static double cd=7654321.0/16777216.0;
static double cm=16777213.0/16777216.0;

void rstart(int i, int j, int k, int l)
{
   int ii, jj, m, wi, wj, wk, wl;
   double s,t;

   wi=i; wj=j; wk=k; wl=l;
   for (ii=0; ii<97; ii++)
   {
      s=0; t=0.5;
      for(jj=1; jj<=24; jj++)
      {
         m=(((wi*wj)%179)*wk)%179;
         wi=wj; wj=wk; wk=m;
         wl=(53*wl+1)%169;
         if ( (wl*m)%64 >= 32 )
            s+=t;
         t*=0.5;
      }
      uu[ii]=s;
   }
}

void t1Init()
{
   Randomize();
   rstart(rand()%178+1,rand()%178+1,rand()%178+1,rand()%169);
}


double t1Jedn(void)
{
   double pom;

   pom=uu[ip-1]-uu[jp-1];
   if (pom < 0.0)
      pom+=1;
   uu[ip-1]=pom;
   ip--;
   if (ip==0)
      ip=97;
   jp--;
   if (jp==0)
      jp=97;
   cc-=cd;
   if (cc<0.0)
      cc+=cm;
   pom-=cc;
   if (pom<0.0) 
      pom+=1;
   return pom;

}

/* Ogolny generator liniowy o okresie rzedu 2^96, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 37. 
   W oryginale blad - omawiany tam generator daje liczby z przedzialu [-0.5,0.5] */
static double a,b,c;

void init_ecng (int ia, int ib, int ic)
{
   a=ia; b=ib; c=ic;
}

void t2Init()
{
   Randomize();
   init_ecng(rand()%100, rand()%100, rand()%100);
}

double t2Jedn()
{
   static int n;
   static double d,x;

   d=(a+b+c)*8192;
   x=fmod(d,4294967291.0);
   a=b; b=c; c=x;
   if (x < (float)2147483648.)
      n= (int) x;
   else
      n= (int)(x-4294967296.);
   return (n*2.3283064365e-10 + 0.5);
}


/* Rozklad bazowy. Zmienna losowa o rozkladzie ciaglym jednostajnym 
   na przedziale [0,1]. Wersja naiwna. */

void t3Init()
{
   Randomize();
}

double t3Jedn ()
{
   return rand()/(double)RAND_MAX;
}


/*
Zmienna losowa o rozkladzie podanym jako argument, okreslona na zbiorze liczb
calkowitych nieujemnych [0..n-1]. Losowanie bez zwracania k elementow.
*/
/*
unsigned ruletka (double* rozklad, unsigned n, unsigned* wynik, unsigned k)
{
   double pom=0, suma, sumar;
   unsigned i,j, traf;

   if (k<n)
      return -1;
   sumar=1;
   for (j=0; j<k; j++)
   {
      pom=jednostajny()*sumar;
      for (i=0, suma=0, traf=n; i<n-1 && traf==n; i++)
      {
         suma+=rozklad[i];
         if (pom < suma)
            traf=i;
      }
      wynik[j]=traf;
      sumar-=rozklad[traf];
   }
   return (j);
}
*/
/*
Zmienna losowa o rozkladzie podanym jako argument, okreslona na zbiorze liczb
calkowitych nieujemnych [0..n-1].  Wersja ze zwracaniem.
*/

unsigned ruletka (double* rozklad, unsigned n)
{
   double pom=0, suma;
   unsigned i;

   pom=jednostajny();
   for (i=0, suma=0; i<n-1; i++)
   {
      suma+=rozklad[i];
      if (pom < suma)
         return i;
   }
   return (n-1);
}

/* 
Zmienna losowa o rozkladzie dyskretnym, rownomiernym na zbiorze liczb 
calkowitych nieujemnych z zakresu [0..n-1]
*/
unsigned rownomierny (unsigned n)
{
   double pom=0, suma;
   unsigned i;

   if (n<=1)
   return 0;
   pom=jednostajny();
   for (i=0, suma=1.0/n; i<n; i++)
      if (pom < suma)
         return i;
      else
         suma +=1.0/n;
   return (n-1);
   
}

/* Generator rozkladu normalnego, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 78, algorytm 3.30. */

double t1Norm()
{
   double u,v, zakres, x, xx;
   

   zakres=sqrt(2/exp(1));
   
   while (1)
   {
      while (!(u=jednostajny()) );
      v=2*zakres*jednostajny()-zakres;
      x=v/u;
      xx=x*x;
      if (xx <= 2*(3-u*(4+u))) return x;
      else if (xx <= 2/u-2*u)
         if (xx <= -4*log(u) ) return x;
   }
   return 0;
   


}
/* Generator rozkladu normalnego, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 42. */

double g(double u)
{
   double t;

   t=sqrt(-2*log(u));
   return (t-
      (((((0.0000453642210148*t)+0.0204231210245)*t+0.342242088547)*t+1)*t+0.322232431088)/
      (((((0.0038560700634*t)+0.103537752850)*t+0.531103462366)*t+0.588581570495)*t+0.0993484626060)
      );
}
double t2Norm()
{
   double u;
   

   u=jednostajny();
   if (u<0.5 && u>1e-20)
      return g(u);
   else
      return (-g(1-u));
   
}
/* Generator rozkladu normalnego Box-Mueller, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 93. */

double t3Norm()
{
   double u,v;
   

   while(!(u=jednostajny()));
   v=jednostajny();
   return (sqrt(-2*log(u))*cos(2*3.141592667597*v));
      
}
/* Zmienna losowa o rozkladzie normalnym standaryzowanym. Opis - por. Pacut
   "Prawdopodobienstwo - Teoria - Modelowanie probabilistyczne w technice",
   WNT Warszawa, 1985, str. 514. */

#define      N 12
double t4Norm ()
{
   unsigned i;
   double suma;

   for (i=0; i<N ; i++)
      suma+=sqrt(12/N)*(jednostajny()-0.5);
   return suma;
}

//double normalny(double mu, double sigma)
//{
//   return (normstd()*sigma+mu);
//}
/*
void normalny(double* X, const double* Mu, const double* Sigma, const unsigned n)
{
   double *Kier, Suma, Ssigma, N;
   unsigned i;

   Kier=(double*) calloc(n, sizeof(double));
   for (i=0, Suma=0; i<n; i++)
   {
      Kier[i]=normstd();
      Suma+=(Kier[i])*(Kier[i]);
   }
   Suma=sqrt(Suma);
   for (i=0; i<n; i++)
      X[i]=Mu[i];

   if (Suma>1e-10)
   {
      for (i=0, Ssigma=0; i<n; i++)
      {
         Kier[i]/=Suma;
         Ssigma+=Sigma[i]*Sigma[i]*Kier[i]*Kier[i];
      }
      Ssigma=sqrt(Ssigma);
      N=Ssigma*normstd();
      for (i=0; i<n; i++)
         X[i]+=N*Kier[i];
   }
   free(Kier);
}
*/      
void normalny(double* X, const double* Mu, const double* Sigma, const unsigned n)
{
   unsigned i;

   for (i=0; i<n; i++)
      X[i]=Mu[i]+Sigma[i]*normstd();
}


/* 
Zmienna losowa o rozkladzie prostokatnym na zbiorze liczb 
rzeczywistych z zakresu [a .. b]
*/
double prostokatny (double a, double b)
{
   return (jednostajny()*(b-a) + a);
}

/* Losowa permutacja zbioru n kolejnych liczb nieujemnych z zakresu [0..n-1]
W tablicy permutacja jest umieszczona wynikowa permutacja
*/

void permutacja (unsigned* permutacja, unsigned n)
{
   unsigned i,j;
   unsigned pom;

   for(i=0; i<n; i++)
      permutacja[i]=i;
   for(i=n-1; i>0; i--)
   {
      j=rownomierny(i);
      if (i!=j)
      {
         pom=permutacja[j];
         permutacja[j]=permutacja[i];
         permutacja[i]=pom;
      }
   }

}


/* 
Zmienna losowa o rozkladzie dwupunktowym, zwracajaca 
1 z pstwem jeden i 0 z pstwem 1-jeden
*/
char dwupunktowy (double jeden)
{
   if (jednostajny() < jeden )
      return 1;
   else
      return 0;
   
}



/*
Zmienna losowa o n-wymiarowym rozkladzie jednostajnym, kazdy z wymiarow 
ma zmiennosc od 0 do 1
*/
void jednNWym (double *wektor, unsigned n)
{
   unsigned i;
   for (i=0; i<n; i++)
      wektor[i]=jednostajny();
}

/*
Zmienna losowa o n-wymiarowym rozkladzie jednostajnym. 
Suma wartosci jest rowna 1.
*/
void normJednNWym (double *wektor, unsigned n)
{
   unsigned i;
   double suma;

   do{
      for (i=0, suma=0; i<n; i++)
         suma+=(wektor[i]=jednostajny());
   }while(suma==0);

   for (i=0; i<n; i++)
      wektor[i]/=suma;
}



/*
Zmienna losowa o n-wymiarowym rozkladzie jednostajnym, kazdy z wymiarow 
ma zmiennosc od 0 do k
*/
void dyskNWym (unsigned *wektor, unsigned n, unsigned k)
{
   unsigned i;
   for (i=0; i<n; i++)
      wektor[i]=rownomierny(k);
}

/* stara wersja
void kombinacja (unsigned* Wynik, unsigned n, unsigned k)
{
   unsigned i,j;
   unsigned Pom;
   unsigned *Permutacja;

   Permutacja = (unsigned*) calloc(n,sizeof(unsigned));

   for(i=0; i<n; i++)
      Permutacja[i]=i;
   for(i=n-1; i>n-k; i--)
   {
      j=rownomierny(i);
      if (i!=j)
      {
    Pom=Permutacja[j];
    Permutacja[j]=Permutacja[i];
    Permutacja[i]=Pom;
      }
   }
   memccpy(Wynik, Permutacja+n-k, k, sizeof(unsigned));
   free(Permutacja);
}
*/

/*
void kombinacja (unsigned* Wynik, unsigned n, unsigned k)
{
   unsigned i,j,l;
   unsigned Pom;
   unsigned *Uzyte;

   Uzyte = (unsigned*) calloc(n,sizeof(unsigned));

   for(i=0; i<n; i++)
      Uzyte[i]=0;
   for (i=0, j=0; i<k; i++)
   {
      l=rownomierny(n)+1;
      do{
   j++;
   while (Uzyte[j]) j=(j+1)%n;
   l--;
      }while (l>0);
      Uzyte[j]=1;
      Wynik[i]=j;
   }
   free(Uzyte);
}

*/
void kombinacja (unsigned* Wynik, unsigned n, unsigned k)
{
   unsigned i,j,l;
   unsigned Pom;
   unsigned *Uzyte;

   Uzyte = (unsigned*) calloc(n,sizeof(unsigned));

   for(i=0; i<n; i++)
      Uzyte[i]=i;
   for (j=0; j<k; j++)
   {
      l=rownomierny(n-j);
     Wynik[j]=Uzyte[l];
     Pom=Uzyte[n-1-j];
      Uzyte[n-1-j]=Uzyte[l];
      Uzyte[l]=Pom;
   }
   free(Uzyte);
}

/* Generator rozkladu Cauchy'ego, opis - por Wieczorkowski, Zielinski
   "Komputerowe generatory liczb losowych", WNT, Warszawa, 1997 str. 91. */

double cauchy()
{
   double u,x;

   x=jednostajny()*2-1;
   u=jednostajny();
   if ((u+0.27324)*(1+x*x) > 1.27324)
      x=jednostajny()*2-1;
   if (jednostajny()>0.5)
      return x;
   else
      if (x!=0)
         return (1/x);
      else
         return MAXDOUBLE;
}

#define LOS_PAR 4
void okreslLosowe (FILE* plikKonf, FILE* plikKontr)
{
   char Wczytanych;
   char *Nazwy[LOS_PAR];
   CZYTFUN czytFun[LOS_PAR];
   void *Adresy[LOS_PAR];
   unsigned *Liczba[LOS_PAR];
   
   Losowe = (LOSOWE )malloc(sizeof(LOSOWES));
   assert(Losowe!=NULL);

   Nazwy[0]="tjednostajny"; czytFun[0]=czytUnsigned; Adresy[0]=&(Losowe->tJednostajny); Liczba[0]=NULL;
   Nazwy[1]="tnormalny"; czytFun[1]=czytUnsigned; Adresy[1]=&(Losowe->tNormalny); Liczba[1]=NULL;
   Nazwy[2]="tinicjacji"; czytFun[2]=czytUnsigned; Adresy[2]=&(Losowe->tInicjacji); Liczba[2]=NULL;
   Nazwy[3]="initziarno"; czytFun[3]=czytUnsigned; Adresy[3]=&(Losowe->InitZiarno); Liczba[3]=NULL;
   
   Wczytanych = wczytajSymbole (plikKonf, plikKontr, Nazwy, czytFun, Adresy, Liczba, LOS_PAR);

/*   assert (Wczytanych==LOS_PAR);*/

}

