#include <math.h>
#include <assert.h>
#include <stdlib.h>

#include "defs.h"
#include "pop.h"
#include "losowe.h"
#include "monitor.h"
#include "stop.h"
#include "zadanie.h"

/*ziarnistosc z jaka przyblizamy dyskrepancje - ile pomiarow w
  proporcji do liczby Zmiennych*/

#define DYSK_DYSK 2000

#ifdef CIAGLE

/*dyskrepancja populacji osobnikow kodowanych rzeczywistoliczbowo*/
double popZroznicowanie(POP Pop)
{
   unsigned i,j,k;
   double *t, *centroid;
   double Wynik, Wynik2;
   double Pom ;
   char Wewnatrz, Wewnatrz2;
   double Objetosc;

   t=(double*)calloc(Zadanie->nZmiennych, sizeof(double));
   centroid=(double*)calloc(Zadanie->nZmiennych, sizeof(double));
   
   for (i=0; i<Zadanie->nZmiennych; i++)
      centroid[i]=0;
   for (k=0; k<Pop->nOsobnik; k++)
      for (i=0; i<Zadanie->nZmiennych; i++)
         centroid[i]+=Pop->Osobniki[k]->Genotyp[i];
   for (i=0; i<Zadanie->nZmiennych; i++)
      centroid[i]/=Pop->nOsobnik;
      
   Wynik=0; Wynik2=0;//Pom2=0;
   for (j=0; j<DYSK_DYSK*Zadanie->nZmiennych; j++)
   {
      Objetosc=1;
      for (i=0; i<Zadanie->nZmiennych; i++)
      {
         t[i]=prostokatny(Zadanie->DolneX[i], Zadanie->GorneX[i]);
         Objetosc*=(t[i]-Zadanie->DolneX[i])/(Zadanie->GorneX[i]-Zadanie->DolneX[i]);
      }
      for (Pom=0,k=0; k<Pop->nOsobnik; k++)
      {
         Wewnatrz=1;
         for (i=0; i<Zadanie->nZmiennych && Wewnatrz; i++)
            if (Pop->Osobniki[k]->Genotyp[i]>t[i])
               Wewnatrz=0;
         if (Wewnatrz)
            Pom+=1;
      }
      Wewnatrz2=1;
      for (i=0; i<Zadanie->nZmiennych && Wewnatrz2; i++)
         if (centroid[i]>t[i])
            Wewnatrz2=0;
      Pom/=Pop->nOsobnik;
      Pom=Objetosc-Pom;
      Wynik+=Pom*Pom;
      if (Wewnatrz2)
         Wynik2+=((Objetosc-1)*(Objetosc-1));
      else
         Wynik2+=Objetosc*Objetosc;
   }
   free (t);
   free (centroid);
   //Wynik=log(sqrt(Wynik/(DYSK_DYSK*Zadanie->nZmiennych))/sqrt(Pom2/(DYSK_DYSK*Zadanie->nZmiennych)));
   Wynik=sqrt(Wynik/(DYSK_DYSK*Zadanie->nZmiennych));
   Wynik2=sqrt(Wynik2/(DYSK_DYSK*Zadanie->nZmiennych));
   Wynik=log(Wynik/Wynik2);

   return Wynik;
}

#else

/* cos na ksztalt entropii populacji osobnikow kodowanych dyskretnie
metoda polega na zliczaniu schematow*/

double entropia (unsigned* Tablica, unsigned N)
{
   unsigned i;
   double Suma, Wynik, Pom;

   for (i=0, Suma=0; i<N; i++)
      Suma+=Tablica[i];
   for (i=0, Wynik=0; i<N; i++)
   {
      if (Suma>0)
      {
         Pom=Tablica[i]/Suma;
         Wynik+=(Pom*log(Pom));
      }
   }
   return Wynik;
}

double popZroznicowanie(POP Pop)
{
   unsigned i,j,k;
   unsigned Istotne;
   unsigned *Znaki;
   unsigned *Schemat;
   double *Entropia;
   double Wynik1, Wynik;

   Schemat=(unsigned*)calloc(Zadanie->nZmiennych, sizeof(unsigned));
   Znaki=(unsigned*)calloc(Zadanie->nLiter, sizeof(unsigned));
   Entropia=(double*)calloc(Zadanie->nZmiennych, sizeof(double));


   for (j=0, Wynik=0; j<DYSK_DYSK*Zadanie->nZmiennych; j++)
   {
      for (i=0; i<Zadanie->nZmiennych; i++)
         Schemat[i]=rownomierny(Zadanie->nLiter+1);
      for (i=0, Wynik1=0, Istotne=0;i<Zadanie->nZmiennych; i++)
      {
         if (Schemat[i]!=Zadanie->nLiter)
         {
            for (k=0; k<Zadanie->nLiter; k++)
               Znaki[k]=0;
            for (k=0; k<Pop->nOsobnik; k++)
               Znaki[Pop->Osobniki[k]->Genotyp[i]]+=1;
            Wynik1+=entropia(Znaki, Zadanie->nLiter);
            Istotne++;
         }
      }

      if (Istotne!=0)
         Wynik+=Wynik1/Istotne; /*entropia istotnych miejsc schematu */
   }
   free (Schemat);
   free (Znaki);
   free (Entropia);
   return Wynik;
}


#endif
MONPOP* inicjujMonPop()
{
   MONPOP* Wynik;
   Wynik=(MONPOP*)malloc(sizeof(MONPOP));
   assert(Wynik!=NULL);
   Wynik->Sredni = osNowy();
   return Wynik;
}

void usunMonPop(MONPOP* Usuwany)
{
   osUsun(&(Usuwany->Sredni));
   free (Usuwany);
}


MONZAD* inicjujMonZad()
{
   MONZAD* Wynik;

   Wynik=(MONZAD*)malloc(sizeof(MONZAD));
   assert(Wynik!=NULL);
   Wynik->Rozwiazanie=NULL;
   Wynik->LPrzystosowan=0;
   Wynik->OkienkoGen = (double*)calloc(Stop->SzerStop, sizeof(double));
   Wynik->OkienkoPrzyst= (double*)calloc(Stop->SzerStop, sizeof(double));
   return Wynik;
}

void usunMonZad (MONZAD* Usuwany)
{
   free(Usuwany->OkienkoGen);
   free(Usuwany->OkienkoPrzyst);
   free(Usuwany);
}
