/*
+--------------------------------------+
|  (C) Jaroslaw Arabas, listopad 1998  |
+--------------------------------------+
*/

#include <assert.h>
#include <stdlib.h>
#include <math.h>
//#include <values.h>
#include "losowe.h"
#include "osobnik.h"
#include "pop.h"
#include "stop.h"
#include "zadanie.h"
#include "monitor.h"

/* sortowanie populacji - wedlug kryterium optymalizacji */

int porownaj (const void *C1, const void *C2)
{
   if ( (*(OSOBNIK*)C2)->Przystosowanie - (*(OSOBNIK*)C1)->Przystosowanie > 0 )
      return 1;
   else
      return -1;
}

void popSortuj (POP Populacja)
{
   qsort(Populacja->Osobniki, Populacja->nOsobnik, sizeof(OSOBNIK), porownaj);
}

/* permutowanie populacji - ustawienie osobnikow w porzadku losowym */
void popPermutuj (POP Populacja)
{
   unsigned *Permut;
   OSOBNIK *Ustawienie;
   unsigned i;

   Permut = (unsigned*)calloc (Populacja->nOsobnik, sizeof(unsigned));
   permutacja (Permut, Populacja->nOsobnik);
   Ustawienie = (OSOBNIK*)calloc (Populacja->nOsobnik, sizeof(OSOBNIK));
   for (i=0; i<Populacja->nOsobnik; i++)
      Ustawienie[i] = Populacja->Osobniki[Permut[i]];
   free (Populacja->Osobniki);
   Populacja->Osobniki=Ustawienie;
   free (Permut);
}



/* ewaluacja populacji - obliczenie wartosci przystosowania dla wszystkich
   osobnikow populacji */
void popEwaluacja (POP Populacja)
{
   unsigned i;

   for (i=0; i<Populacja->nOsobnik; i++)
   {
      if (wKostkowych(Populacja->Osobniki[i]))
      {
         fPrzystosowania (Populacja->Osobniki[i]);
         Zadanie->Monitor->LPrzystosowan++;
      }
      else
         Populacja->Osobniki[i]->Przystosowanie=SMIERC;
   }
   

}

POP popNowa(unsigned nOsobnik)
{
   POP Wynik;

   Wynik=(POP)malloc(sizeof(POPS));
   Wynik->nOsobnik=nOsobnik;
   Wynik->Osobniki=(OSOBNIK*)calloc(nOsobnik, sizeof(OSOBNIK));
   Wynik->Monitor = inicjujMonPop();
   return Wynik;
}


/* inicjacja populacji - wypelnienie jej nowymi osobnikami */
void popOsobniki(POP Populacja)
{
   unsigned i;

   for(i=0; i<Populacja->nOsobnik; i++)
      Populacja->Osobniki[i]=osNowy();

}

/* inicjacja populacji - wypelnienie jej nowymi osobnikami */
void popInic (POP Populacja)
{
   unsigned i;
   TGEN **Genotypy;

   Genotypy=(TGEN**)calloc(Populacja->nOsobnik, sizeof(TGEN*));
   assert(Genotypy!=NULL);
   for(i=0; i<Populacja->nOsobnik; i++)
      Genotypy[i]=Populacja->Osobniki[i]->Genotyp;
   zadInicjuj(Genotypy, Populacja->nOsobnik);
   free(Genotypy);
}

/* usuniecie osobnikow z populacji */
void popCzysc(POP Populacja)
{
   unsigned i;

   for (i=0;i<Populacja->nOsobnik; i++)
      if (Populacja->Osobniki[i] !=NULL)
         osUsun(&(Populacja->Osobniki[i]));
}


void popUsun(POP Populacja)
{
   popCzysc(Populacja);
   free(Populacja->Osobniki);
   usunMonPop(Populacja->Monitor);
   free(Populacja);
}

void popMonitor(POP Pop)
{
   unsigned i,j;
   unsigned Zywi;

   /*Wykonanie statystyk populacji */
   Pop->Monitor->Maksimum=0;
   Pop->Monitor->Minimum=0;
   Pop->Monitor->Sredni->Przystosowanie=0;
   Pop->Monitor->Zroznicowanie=0;
#ifdef CIAGLE
   for (j=0; j<Pop->Monitor->Sredni->nGenow; j++)
      Pop->Monitor->Sredni->Genotyp[j]=0;
#endif
   for (j=0; j<Pop->Monitor->Sredni->nPar; j++)
      Pop->Monitor->Sredni->Parametry[j]=0;
   for (i=0, Zywi=0; i<Pop->nOsobnik; i++)
   {

#ifdef CIAGLE
      for (j=0; j<Pop->Monitor->Sredni->nGenow; j++)
         Pop->Monitor->Sredni->Genotyp[j]+=Pop->Osobniki[i]->Genotyp[j];
#endif

         for (j=0; j<Pop->Monitor->Sredni->nPar; j++)
         Pop->Monitor->Sredni->Parametry[j]+=Pop->Osobniki[i]->Parametry[j];

      if (Pop->Osobniki[i]->Przystosowanie!=SMIERC)
      {
         if (Pop->Osobniki[i]->Przystosowanie > Pop->Osobniki[Pop->Monitor->Maksimum]->Przystosowanie)
            Pop->Monitor->Maksimum=i;
         if (Pop->Osobniki[i]->Przystosowanie < Pop->Osobniki[Pop->Monitor->Minimum]->Przystosowanie)
            Pop->Monitor->Minimum=i;
         Pop->Monitor->Sredni->Przystosowanie+=Pop->Osobniki[i]->Przystosowanie;
         Zywi++;
	 Pop->Osobniki[i]->Wiek++;
      }

   }

#ifdef CIAGLE
   for (j=0; j<Pop->Monitor->Sredni->nGenow; j++)
      Pop->Monitor->Sredni->Genotyp[j]/=Pop->nOsobnik;
#endif

   for (j=0; j<Pop->Monitor->Sredni->nPar; j++)
      Pop->Monitor->Sredni->Parametry[j]/=Pop->nOsobnik;
   

   if (Zywi!=0)
   {
      Pop->Monitor->Sredni->Przystosowanie/=Zywi;
#ifdef ZROZNICOWANIE
      Pop->Monitor->Zroznicowanie=popZroznicowanie(Pop);
#endif
   }
   else
      Pop->Monitor->Sredni->Przystosowanie=SMIERC;
}

void zadMonitor(unsigned t, POP Pt)
{
   unsigned i;
   /* Wykonanie statystyk algorytmu */
   Zadanie->Monitor->Generacja=t;
   for (i=1; i<Stop->SzerStop; i++)
      Zadanie->Monitor->OkienkoGen[i-1]=Zadanie->Monitor->OkienkoGen[i];

   
   if (Zadanie->Monitor->OkienkoGen[Stop->SzerStop-1]< 
      Pt->Osobniki[Pt->Monitor->Maksimum]->Przystosowanie  || Zadanie->Monitor->Generacja ==0)
   {
      if (Zadanie->Monitor->Rozwiazanie!=NULL)
         osUsun(&(Zadanie->Monitor->Rozwiazanie));
      Zadanie->Monitor->Rozwiazanie=osKopiuj2(Pt->Osobniki[Pt->Monitor->Maksimum]);
      Zadanie->Monitor->OkienkoGen[Stop->SzerStop-1]=Zadanie->Monitor->Rozwiazanie->Przystosowanie;
   }
   else
      Zadanie->Monitor->Rozwiazanie->Wiek++;

   
}





