/*
+--------------------------------------+
|  (C) Jaroslaw Arabas, luty 1999      |
+--------------------------------------+
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "pop.h"
#include "losowe.h"
#include "algorytm.h"
#include "raporty.h"
#include "selekcja.h"
#include "wewy.h"

SELEKCJA Selekcja;

char rPropMod (POP Pt, unsigned *Wybrani, unsigned nWybranych );
char rTurnZwr (POP Pt, unsigned *Wybrani, unsigned nWybranych );
char rRang (POP Pt, unsigned *Wybrani, unsigned nWybranych );
char rProsta (POP Pt, unsigned *Wybrani, unsigned nWybranych );
char rPropStd (POP Pt, unsigned *Wybrani, unsigned nWybranych );
char rTurnBZ (POP Pt, unsigned *Wybrani, unsigned nWybranych );

char sProsta (POP Pnowa, POP Pt, POP Ot);
char sElitaStd (POP Pnowa, POP Pt, POP Ot);
char sElitaCZ (POP Pnowa, POP Pt, POP Ot);

double fRWykl (unsigned Ranga, unsigned MaxRanga);
double fRProg (unsigned Ranga, unsigned MaxRanga);

void rangiWartosc(unsigned nOsobnik, unsigned* Sort,unsigned* Rangi,unsigned* MaxRanga, OSOBNIK *Osobniki);
void rangiNumer(unsigned nOsobnik, unsigned* Sort,unsigned* Rangi,unsigned* MaxRanga, OSOBNIK *Osobniki);
char reprodukcja(POP Pt, unsigned *Rodzice, const unsigned nRodzicow)
{
   switch (Selekcja->tReprodukcji)
   {
   case 1:  return rPropStd(Pt, Rodzice, nRodzicow);
   case 2:  return rPropMod(Pt, Rodzice, nRodzicow);
   case 3:  return rTurnZwr(Pt, Rodzice, nRodzicow);
   case 4:  return rTurnBZ(Pt, Rodzice, nRodzicow);
   case 5:  return rRang(Pt, Rodzice, nRodzicow);
   case 6:  return rProsta(Pt, Rodzice, nRodzicow);
   default: return 1;
   }
	   
}


char sukcesja (POP Pnowa, POP Pt, POP Ot)
{
   switch (Selekcja->tSukcesji)
   {
   case 1:  return sProsta (Pnowa, Pt, Ot);
   case 2:  return sElitaStd (Pnowa, Pt, Ot);
   case 3:  return sElitaCZ (Pnowa, Pt, Ot);
   default: return 1;
   }

}

double fRang (unsigned Ranga, unsigned MaxRanga)
{
   switch (Selekcja->fRang)
   {
   case 1: return fRWykl (Ranga, MaxRanga);
   case 2: return fRProg (Ranga, MaxRanga);
   default: return 1;
   }
}
/* Reprodukcja proporcjonalna (ruletkowa) z przesuwaniem wartosci 
   funkcji przystosowania. Dzieki temu, wartosci te sa zawsze nie mniejsze
   od zera */

char rPropMod (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   double *RozkladSel;                /*rozklad p-stwa selekcji*/
   unsigned i,j;
   double SumaPrzyst;

/* utworzenie dystrybuanty rozkladu p-stwa selekcji                  */

   RozkladSel =(double*) calloc (Pt->nOsobnik, sizeof(double));
   
   SumaPrzyst=(Pt->Monitor->Sredni->Przystosowanie-
      Pt->Osobniki[Pt->Monitor->Minimum]->Przystosowanie )*Pt->nOsobnik;
   if (SumaPrzyst!=0)
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = (Pt->Osobniki[i]->Przystosowanie - Pt->Osobniki[Pt->Monitor->Minimum]->Przystosowanie) / SumaPrzyst;
   else
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = 1;

   
   if (Selekcja->BezZwracania)
/* WARIANT BEZ ZWRACANIA   */
   {
      for(i=0; i<nWybranych; i++)
      {
         Wybrani[i] = ruletka (RozkladSel, Pt->nOsobnik);
         for (j=0; j<Pt->nOsobnik; j++)
            RozkladSel[j]*=( SumaPrzyst/(SumaPrzyst-RozkladSel[Wybrani[i]]) );
         RozkladSel[Wybrani[i]]=0;
      }
   }
   else
/* WARIANT ZE ZWRACANIEM     */
   {
      for(i=0; i<nWybranych; i++)
         Wybrani[i] = ruletka (RozkladSel, Pt->nOsobnik);
   }

/* zwolnienie zbednej pamieci                                        */
   free (RozkladSel);

   return 0;
}
unsigned zwyciezca (OSOBNIK* Konkurenci, unsigned* Turniej, unsigned nTurnieju)
{
   unsigned i, Zwyciezca;

   for(i=1, Zwyciezca=Turniej[0]; i<nTurnieju; i++)
      if (Konkurenci[Turniej[i]]->Przystosowanie > Konkurenci[Zwyciezca]->Przystosowanie)
         Zwyciezca=Turniej[i];
   return Zwyciezca;
}


/* Reprodukcja turniejowa. Do turnieju wchodza osobniki ze zwracaniem, 
moze sie wiec zdarzyc, ze jeden osobnik zostanie wybrany wiecej niz
jednokrotnie. Wielkosc turnieju jest zdefiniowana jako zmienna globalna
*/

char rTurnZwr (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   unsigned Zwyciezca;
   unsigned *Turniej;
   unsigned i,j, k;


/* utworzenie miejsca na turniej                                     */
   Turniej = (unsigned*)calloc (Selekcja->nTurnieju, sizeof(unsigned));
   
   if(Selekcja->BezZwracania)
	   
/* WARIANT BEZ ZWRACANIA                 */
   {

      double *RozkladSel;                /*rozklad p-stwa selekcji*/
      unsigned LiczbaOs;
/* utworzenie dystrybuanty rozkladu p-stwa selekcji                  */

      RozkladSel =(double*) calloc (Pt->nOsobnik, sizeof(double));
   
      LiczbaOs=Pt->nOsobnik;
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = 1.0/LiczbaOs;
      for(i=0; i<nWybranych; i++)
      {
         do{
            for (j=0; j<Selekcja->nTurnieju; j++)
               Turniej [j] = ruletka (RozkladSel, Pt->nOsobnik);
      
            Zwyciezca=zwyciezca (Pt->Osobniki, Turniej, Selekcja->nTurnieju); 

         }while(Pt->Osobniki[Zwyciezca]->Przystosowanie == SMIERC);
         Wybrani[i]=Zwyciezca;
         for (k=0; k<Pt->nOsobnik; k++)
            RozkladSel[k]*=((double)LiczbaOs /(LiczbaOs-1) );
         RozkladSel[Zwyciezca]=0;
	 LiczbaOs--;
      }
      free (RozkladSel);
   }
   else
/* WARIANT ZE ZWRACANIEM                 */
   {
      for(i=0; i<nWybranych; i++)
      {
         do{
            for (j=0; j<Selekcja->nTurnieju; j++)
               Turniej [j] = rownomierny (Pt->nOsobnik);
      
            Zwyciezca=zwyciezca (Pt->Osobniki, Turniej, Selekcja->nTurnieju); 

         }while(Pt->Osobniki[Zwyciezca]->Przystosowanie == SMIERC);
         Wybrani[i]=Zwyciezca;
      }
   }
	   

/* czynnosci porzadkowe */
   free (Turniej);

   return 0;
}
      

#define EPS_TOL 0.0
// nieefektywna metoda sortowania
void wyznaczRangi (unsigned *Rangi, unsigned *MaxRanga, 
		   unsigned nOsobnik, OSOBNIK *Osobniki)
{
   unsigned i,j;
//   double Tolerancja;
   unsigned Pom;
   unsigned *Sort;
   unsigned Ranga;

   Sort=(unsigned*)calloc(nOsobnik, sizeof(unsigned));

   //sortowanie numerow osobnikow - metoda brutalna
   for (i=0; i<nOsobnik; i++)
      Sort[i]=i;
   for (i=0; i<nOsobnik-1; i++)
      for(j=i+1; j<nOsobnik; j++)
      {
         if (Osobniki[Sort[i]]->Przystosowanie < Osobniki[Sort[j]]->Przystosowanie)
         {
            Pom=Sort[i];
            Sort[i]=Sort[j];
            Sort[j]=Pom;
         }
         
      }

   switch(Selekcja->tRang)
   {
	   case 1: rangiWartosc(nOsobnik, Sort, Rangi, MaxRanga, Osobniki);
		   break;
	   case 2: rangiNumer(nOsobnik, Sort, Rangi, MaxRanga, Osobniki);
		   break;
	   default: rangiNumer(nOsobnik, Sort, Rangi, MaxRanga, Osobniki);
		   break;
   }

   free(Sort);
      
}
void rangiWartosc (unsigned nOsobnik, unsigned* Sort, unsigned *Rangi, 
                   unsigned *MaxRanga, OSOBNIK *Osobniki)

{
   unsigned i, Ranga;
   double Prototyp;
   
   //przypisanie rang osobnikom
   Rangi[Sort[0]]=Ranga=1;
   for (i=1, Prototyp=Osobniki[Sort[0]]->Przystosowanie; i<nOsobnik; i++)
   {
      if (Prototyp-Osobniki[Sort[i]]->Przystosowanie <= EPS_TOL)
         Rangi[Sort[i]]=Ranga;
      else
      {
         Prototyp=Osobniki[Sort[i]]->Przystosowanie;
         Rangi[Sort[i]]=++Ranga;
      }
   }
   *MaxRanga=Ranga;
}
void rangiNumer (unsigned nOsobnik, unsigned* Sort, unsigned *Rangi, 
                 unsigned *MaxRanga, OSOBNIK *Osobniki)
{
   unsigned i;
   for (i=0; i<nOsobnik; i++)
   {
      Rangi[Sort[i]]=i+1;
   }
   *MaxRanga=(nOsobnik);
}
	

/* funkcja rang - wykladnicza*/
double fRWykl(unsigned Ranga, unsigned MaxRanga)
{
   double w;

   return (pow((MaxRanga - Ranga),Selekcja->fRWykladnik));
}

double fRProg (unsigned Ranga, unsigned MaxRanga)
{
   unsigned mu;

   mu=MaxRanga * Selekcja->obciecie;
   if (mu <1)
      mu=1;
   if (Ranga <= mu)
      return (1.0/mu);
   else
      return 0;
}

void wyznaczRozklad(double *Rozklad, unsigned *Rangi, unsigned MaxRanga, 
		    unsigned nOsobnik)
{
   unsigned i;
   double Suma;
   double Mnoznik;
   

   for (i=0,Suma=0; i<nOsobnik; i++)
   {
      Rozklad[i]=fRang(Rangi[i], MaxRanga);
      Suma+=Rozklad[i];
   }
//   Mnoznik=nOsobnik/Suma;
   for (i=0; i<nOsobnik; i++)
      Rozklad[i]= Selekcja->rStala + (1-Selekcja->rStala)/Suma* Rozklad[i];
   //      Rozklad[i]*=Mnoznik;
}

/* Reprodukcja rangowa niedeterministyczna. 
*/

char rRang (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   unsigned *Rangi;
   double *Rozklad;
   unsigned i,j;
   unsigned MaxRanga;

/* przygotowanie tablicy przekodowania rang na liczbe kopii             */
   Rangi = (unsigned*)calloc (Pt->nOsobnik, sizeof(unsigned));
   Rozklad = (double*)calloc (Pt->nOsobnik, sizeof(double));

   wyznaczRangi(Rangi, &MaxRanga, Pt->nOsobnik, Pt->Osobniki);

//   MaxRanga=Pt->nOsobnik; //atrapa
   wyznaczRozklad(Rozklad, Rangi, MaxRanga, Pt->nOsobnik);

/* wlasciwa selekcja rangowa                                            */

   for (i=0; i<nWybranych ; i++)
   {
      j=ruletka(Rozklad, Pt->nOsobnik);
      Wybrani[i] = j;
   }

/* czynnosci porzadkowe */
   free (Rozklad);
   free (Rangi);

   return 0;


}
     
/* Reprodukcja prosta - nie ma zroznicowania prawdopodobienstw. */

char rProsta (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   unsigned i;

/* reprodukcja przez wielokrotny wybor ze zwracaniem i kopiowanie    */
/* wylosowanego osobnika - wybor z jednakowym prawdopodobienstwem    */

   for(i=0; i<nWybranych; i++)
      Wybrani[i] = rownomierny(Pt->nOsobnik);

   return 0;
}

/* Reprodukcja proporcjonalna (ruletkowa) bez przesuwania wartosci 
   funkcji przystosowania. */
   
 
char rPropStd (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   double *RozkladSel;                /*rozklad p-stwa selekcji*/
   unsigned i,j;
   double SumaPrzyst;

/* utworzenie dystrybuanty rozkladu p-stwa selekcji                  */

   RozkladSel =(double*) calloc (Pt->nOsobnik, sizeof(double));
   
   SumaPrzyst=(Pt->Monitor->Sredni->Przystosowanie)*Pt->nOsobnik;
   if (SumaPrzyst!=0)
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = (Pt->Osobniki[i]->Przystosowanie ) / SumaPrzyst;
   else
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = 1;

   
   if (Selekcja->BezZwracania)
	   
/* WARIANT BEZ ZWRACANIA    */

   {
      for(i=0; i<nWybranych; i++)
      {
         Wybrani[i] = ruletka (RozkladSel, Pt->nOsobnik);
         for (j=0; j<Pt->nOsobnik; j++)
            RozkladSel[j]*=( SumaPrzyst/(SumaPrzyst-RozkladSel[Wybrani[i]]) );
         RozkladSel[Wybrani[i]]=0;
      }
   }
   else
/* WARIANT ZE ZWRACANIEM    */

   {
      for(i=0; i<nWybranych; i++)
         Wybrani[i] = ruletka (RozkladSel, Pt->nOsobnik);
   }

/* zwolnienie zbednej pamieci                                        */
   free (RozkladSel);

   return 0;
}

/* Reprodukcja turniejowa. Do turnieju wchodza osobniki bez zwracania, 
Wielkosc turnieju jest zdefiniowana jako zmienna globalna.
*/

char rTurnBZ (POP Pt, unsigned *Wybrani, unsigned nWybranych )
{
   unsigned Zwyciezca;
   unsigned *Turniej;
   unsigned i,j,k;

/* utworzenie miejsca na turniej                                     */
   Turniej = (unsigned*)calloc (Selekcja->nTurnieju, sizeof(unsigned));
   

   if (Selekcja->BezZwracania)
/* WARIANT BEZ ZWRACANIA                 */
   {
      double *RozkladSel, *RozkladSel2;            /*rozklad p-stwa selekcji*/
      unsigned LiczbaOs;

   /* utworzenie dystrybuanty rozkladu p-stwa selekcji                  */

      RozkladSel =(double*) calloc (Pt->nOsobnik, sizeof(double));
      RozkladSel2 =(double*) calloc (Pt->nOsobnik, sizeof(double));
   
      LiczbaOs=Pt->nOsobnik;
      for(i=0; i<Pt->nOsobnik; i++)
         RozkladSel[i] = 1.0/LiczbaOs;
      for(i=0; i<nWybranych; i++)
      {
         do{
	    for(k=0; k<Pt->nOsobnik; k++)
	 	 RozkladSel2[k]=RozkladSel[k];
            for (j=0; j<Selekcja->nTurnieju; j++)
            {
               Turniej [j] = ruletka (RozkladSel2, Pt->nOsobnik);
               for (k=0; k<Pt->nOsobnik; k++)
                  RozkladSel2[k]*=((double)(LiczbaOs-j) /(LiczbaOs-j-1) );
               RozkladSel2[Turniej[j]]=0;
            }
      
            Zwyciezca=zwyciezca (Pt->Osobniki, Turniej, Selekcja->nTurnieju); 
         }while(Pt->Osobniki[Zwyciezca]->Przystosowanie == SMIERC);
         Wybrani[i]=Zwyciezca;
         for (k=0; k<Pt->nOsobnik; k++)
            RozkladSel[k]*=((double)LiczbaOs /(LiczbaOs-1) );
         RozkladSel[Zwyciezca]=0;
         LiczbaOs--;
      }
      free (RozkladSel);
      free (RozkladSel2);
   }
   else
/* WARIANT ZE ZWRACANIEM*/
   {
      for(i=0; i<nWybranych; i++)
      {
         do{
            kombinacja(Turniej, Pt->nOsobnik, Selekcja->nTurnieju);
      
            Zwyciezca=zwyciezca (Pt->Osobniki, Turniej, Selekcja->nTurnieju); 
         }while(Pt->Osobniki[Zwyciezca]->Przystosowanie == SMIERC);
         Wybrani[i]=Zwyciezca;
      }
   }


/* czynnosci porzadkowe */
   free (Turniej);


   return 0;

}
      

/* Sukcesja nieelitarna --- z calkowitym zastepowaniem*/

char sProsta (POP Pnowa, POP Pt, POP Ot)
{
   unsigned i;

   for (i=0; i<Pnowa->nOsobnik; i++)
      Pnowa->Osobniki [i] = osPrzesun (&(Ot->Osobniki[i]));
   return 0;
}

/* Sukcesja elitarna */
char sElitaStd (POP Pnowa, POP Pt, POP Ot)
{
   unsigned i, j, k;

   popSortuj (Pt);
   popSortuj (Ot);
   
/* Wybieranie najlepszych osobnikow ze zlaczenia Pt i Ot do nastepnej 
populacji bazowej. Pierwszenstwo (przy jednakowych wartosciach funkcji
przystosowania) maja nowo wygenerowane Osobniki z Pt
*/

   j=0; k=0;
   for (i=0; i<Pnowa->nOsobnik; i++)
   {  
/* jesli nie ma juz osobnikow w Pt albo Ot - koniec petli*/
      if (k>=Ot->nOsobnik || j>=Selekcja->nElity*Algorytm->nPop) break;

/* wybor lepszego do skopiowania */
      if (Ot->Osobniki[k]->Przystosowanie > Pt->Osobniki[j]->Przystosowanie)
         Pnowa->Osobniki [i] = osPrzesun (&(Ot->Osobniki[k++]));
      else if (Ot->Osobniki[k]->Przystosowanie == Pt->Osobniki[j]->Przystosowanie)
         Pnowa->Osobniki [i] = osPrzesun (&(Pt->Osobniki[j++]));
      else
         Pnowa->Osobniki [i] = osPrzesun (&(Pt->Osobniki[j++]));
   }

/* uzupelnienie ewentualnych brakow osobnikami z tej populacji, ktora 
nie zostala wyczerpana */
   if (i<Pnowa->nOsobnik)
   {
      if (k>=Ot->nOsobnik)
         for (; i<Pnowa->nOsobnik; i++)
            Pnowa->Osobniki [i] = osPrzesun (&(Pt->Osobniki[j++]));
      else
         for (; i<Pnowa->nOsobnik; i++)
            Pnowa->Osobniki [i] = osPrzesun (&(Ot->Osobniki[k++]));
   }

/* czynnosci porzadkowe */
   return 0;
}


/* Sukcesja elitarna z maksymalnym  czasem zycia */
char sElitaCZ (POP Pnowa, POP Pt, POP Ot)
{
   
   unsigned i, j, k;

   for (i=0 ; i<Pt->nOsobnik ; i++)
      if (Pt->Osobniki[i]->Wiek > Selekcja->maxWiek)
         Pt->Osobniki[i]->Przystosowanie = SMIERC;
   
   return sElitaStd (Pnowa, Pt, Ot);
}

#define SEL_PAR 11
void okreslSelekcje(FILE *plikKonf, FILE *plikKontr)
{                
   char Wczytanych;
   char *Nazwy[SEL_PAR];
   CZYTFUN czytFun[SEL_PAR];
   void *Adresy[SEL_PAR];
   unsigned *Liczba[SEL_PAR];

   Selekcja = (SELEKCJA )malloc(sizeof(SELEKCJAS));
   assert(Selekcja!=NULL);

   Nazwy[0]="treprodukcji"; czytFun[0]=czytUnsigned; Adresy[0]=&(Selekcja->tReprodukcji); Liczba[0]=NULL;
   Nazwy[1]="tsukcesji"; czytFun[1]=czytUnsigned; Adresy[1]=&(Selekcja->tSukcesji); Liczba[1]=NULL;
   Nazwy[2]="nturnieju"; czytFun[2]=czytUnsigned; Adresy[2]=&(Selekcja->nTurnieju); Liczba[2]=NULL;
   Nazwy[3]="nelity"; czytFun[3]=czytDouble; Adresy[3]=&(Selekcja->nElity); Liczba[3]=NULL;
   Nazwy[4]="trang"; czytFun[4]=czytUnsigned; Adresy[4]=&(Selekcja->tRang); Liczba[4]=NULL;
   Nazwy[5]="maxwiek"; czytFun[5]=czytUnsigned; Adresy[5]=&(Selekcja->maxWiek); Liczba[5]=NULL;
   Nazwy[6]="obciecie"; czytFun[6]=czytDouble; Adresy[6]=&(Selekcja->obciecie); Liczba[6]=NULL;
   Nazwy[7]="frwykladnik"; czytFun[7]=czytDouble ; Adresy[7]=&(Selekcja->fRWykladnik); Liczba[7]=NULL;
   Nazwy[8]="bezzwracania"; czytFun[8]=czytUnsigned ; Adresy[8]=&(Selekcja->BezZwracania); Liczba[8]=NULL;
   Nazwy[9]="frang"; czytFun[9]=czytUnsigned ; Adresy[9]=&(Selekcja->fRang); Liczba[9]=NULL;
   Nazwy[10]="rstala"; czytFun[10]=czytDouble ; Adresy[10]=&(Selekcja->rStala); Liczba[10]=NULL;

   Wczytanych = wczytajSymbole (plikKonf, plikKontr, Nazwy, czytFun, Adresy, Liczba, SEL_PAR);

/*   assert (Wczytanych==SEL_PAR);*/

}

